import obspy

from decode_mseed import (
    decode_ieee_float, decode_ieee_double, decode_steim, decode_int16,
    decode_int24, decode_int32,
)
from help_do_stuff import (
    FixedHeader, Blockette1000, get_data_endianness, Unpacker,
    get_record_metadata, get_header_endianness, RecordMetadata,
    EncodingFormat,
)


class RecordReader:
    def __init__(self, file):
        self.file = file
        self.fixed_header: FixedHeader = None
        self.blockette_1000: Blockette1000 = None
        # Utility object that helps unpack byte string in fixed header.
        # Separate from the one for data in case the header has a different
        # byte order.
        # TODO: change blockettes to use this unpacker as well.
        self.fixed_header_unpacker: Unpacker = Unpacker()
        self.data_unpacker: Unpacker = Unpacker()
        self.record_metadata: RecordMetadata = None

    def read_fixed_header(self):
        byte_counts = [6, 1, 1, 5, 2, 3, 2, 10, 2, 2, 2, 1, 1, 1, 1, 4, 2, 2]

        fixed_header_sections_values = []
        for byte_count in byte_counts:
            fixed_header_sections_values.append(self.file.read(byte_count))
        self.fixed_header = FixedHeader(*fixed_header_sections_values)

    def read_blockette_500(self):
        """
        Read blockette 500. Only read more precise start time. Assumes that the
        file pointer is at the start of the blockette.
        # TODO: find out what VCO correction is.
        """
        self.file.seek(18, 1)
        start_time_microsecond = self.file.read(1)
        start_time_microsecond = self.data_unpacker.unpack(
            'b', start_time_microsecond
        )[0]
        # Convert from microsecond to second so that UTCDateTime can handle it.
        start_time_microsecond /= (10 ** 6)
        self.record_metadata.start_time += start_time_microsecond
        self.file.seek(181, 1)

    def read_blockette_1000(self):
        blockette_1000_section_lengths = [2, 2, 1, 1, 1, 1]
        blockette_1000_values = []
        for section_length in blockette_1000_section_lengths:
            blockette_1000_values.append(self.file.read(section_length))

        self.blockette_1000 = Blockette1000(*blockette_1000_values)

    def read_blockette_1001(self):
        """
        Read blockette 1001. The only valuable thing in this blockette is the
        more precise start time. Assumes that the file pointer is at the start
        of the blockette.
        """
        self.file.read(5)
        start_time_microsecond = self.file.read(1)
        start_time_microsecond = self.data_unpacker.unpack(
            'b', start_time_microsecond
        )[0]
        # Convert from microsecond to second so that UTCDateTime can handle it.
        start_time_microsecond /= (10**6)
        self.record_metadata.start_time += start_time_microsecond
        self.file.read(2)

    def apply_time_correction(self):
        # format() is used here instead of bin() because we need to pad the
        # resulting bit string with 0 to the left.
        activity_flags = format(
            self.fixed_header_unpacker.unpack(
                'B', self.fixed_header.activity_flags)[0],
            '0>8b'
        )
        is_time_correction_applied = int(activity_flags[1])
        if is_time_correction_applied:
            return

        time_correction = self.fixed_header_unpacker.unpack(
            'L', self.fixed_header.time_correction
        )[0]
        # We need to convert the unit from 0.0001 seconds to seconds
        time_correction *= 0.0001
        self.record_metadata.start_time += time_correction

    def read_blockettes(self):
        blockette_count = self.fixed_header_unpacker.unpack(
            'B', self.fixed_header.blockette_count
        )[0]
        for i in range(1, blockette_count):
            # All blockettes store their type in the first two bytes, so we
            # read that to determine what to do
            next_blockette_type = self.file.read(2)
            # Move file pointer back to start of blockette
            self.file.seek(-2, 1)
            next_blockette_type = self.data_unpacker.unpack(
                'H', next_blockette_type
            )[0]
            if next_blockette_type not in (1000, 1001, 500):
                print('We currently only handle blockettes 500, 1000, and'
                      '1001.')
                continue
            if next_blockette_type == 1001:
                self.read_blockette_1001()
            elif next_blockette_type == 500:
                self.read_blockette_500()

    def get_first_data_point(self):
        record_start = self.file.tell()
        self.read_fixed_header()
        self.read_blockette_1000()

        header_endianness = get_header_endianness(self.fixed_header)
        if header_endianness == 'little':
            self.fixed_header_unpacker.byte_order_char = '<'
        else:
            self.fixed_header_unpacker.byte_order_char = '>'

        data_endianness = get_data_endianness(self.blockette_1000)
        if data_endianness == 'little':
            self.data_unpacker.byte_order_char = '<'
        else:
            self.data_unpacker.byte_order_char = '>'

        self.record_metadata = get_record_metadata(self.fixed_header,
                                                   self.fixed_header_unpacker)

        self.apply_time_correction()
        self.read_blockettes()
        data_start = self.fixed_header_unpacker.unpack('H', self.fixed_header.data_offset)[0]
        # It is most convenient to seek to the start of the record, then seek
        # to the start of the data. Any other way would involve some
        # calculation to determine which point to seek to.
        self.file.seek(record_start)
        self.file.seek(data_start, 1)

        encoding_format = self.blockette_1000.encoding_format
        encoding_format = self.data_unpacker.unpack('b', encoding_format)[0]
        encoding_format = EncodingFormat(encoding_format)

        # Currently, we are extracting only the first data point in each
        # record. The smallest possible amount of bytes we can extract while
        # guaranteeing that we get the first data point in the record is 8,
        # with Steim encodings and IEEE double precision float needing to use
        # the whole buffer.
        buffer = self.file.read(8)
        encoding_to_decoder = {
            EncodingFormat.INT_16_BIT: decode_int16,
            EncodingFormat.INT_24_BIT: decode_int24,
            EncodingFormat.INT_32_BIT: decode_int32,
            EncodingFormat.IEEE_FLOAT_32_BIT: decode_ieee_float,
            EncodingFormat.IEEE_FLOAT_64_BIT: decode_ieee_double,
            EncodingFormat.STEIM_1: decode_steim,
            EncodingFormat.STEIM_2: decode_steim,
        }
        first_data_point = encoding_to_decoder[encoding_format](buffer, self.data_unpacker)

        return first_data_point


if __name__ == '__main__':
    # numpy.set_printoptions(threshold=sys.maxsize)
    file_path = '/Users/kle/PycharmProjects/sohstationviewer/tests/test_data/potato.sdr/XX.KC01..VCO.D.2020.145'
    file = open(file_path, 'rb')
    stream = obspy.read(file_path)

    trace = []
    while 1:
        # We know that end of file is reached when read() returns an empty
        # string.
        is_eof = (file.read(1) == b'')
        if is_eof:
            break
        # We need to move the file pointer back to its position before we do
        # the end of file check. Otherwise, we would be off by one byte for all
        # the reads afterward.
        file.seek(-1, 1)

        # We save the start of the current record so that after we are done
        # reading the record, we can move back. This makes moving to the next
        # record a lot easier, seeing as we can simply move the file pointer
        # a distance the size of the current record.
        current_record_start = file.tell()

        reader = RecordReader(file)
        trace.append(reader.get_first_data_point())
        sample_count = reader.record_metadata.sample_count
        sample_rate = reader.record_metadata.sample_rate
        record_time_taken = sample_count / sample_rate
        record_end_time = reader.record_metadata.start_time + record_time_taken

        # MSEED stores the size of a data record as an exponent of a power of
        # two, so we have to convert that to actual size before doing anything
        # else.
        record_length_exp = reader.data_unpacker.unpack(
            'B', reader.blockette_1000.record_length
        )[0]
        record_size = 2 ** record_length_exp

        file.seek(current_record_start)
        file.seek(record_size, 1)
    print(trace)
